"""Un/subscribe and consume AutoKitteh connection events."""

from datetime import timedelta
from uuid import uuid4

from .attr_dict import AttrDict


def subscribe(source: str, filter: str = "") -> str:
    """Subscribe to events on connection. Optional filter is a CEL expression.

    Works both in durable and nondurable sessions.
    """
    # Dummy implementation for local development.
    return f"sig_{uuid4().hex}"


def unsubscribe(subscription_id: str) -> None:
    """Unsubscribe from events.

    Works both in durable and nondurable sessions.
    """
    # Dummy implementation for local development.
    pass


def next_event(
    subscription_id: str | list[str], *, timeout: timedelta | int | float = None
) -> AttrDict:
    """Get the next event from the subscription(s).

    If timeout is not None and there are no new events after timeout, this function will
    return None.

    Works both in durable and nondurable sessions.
    """
    # Dummy implementation for local development.
    return AttrDict()


def start(
    loc: str,
    data: dict | None = None,
    memo: dict | None = None,
    project: str = "",
) -> str:
    """Start a new session.

    Works both in durable and nondurable sessions.

    Sessions started this way will have the same durability setting as the parent session.
    """
    # Dummy implementation for local development.
    return f"ses_{uuid4().hex}"



from pathlib import Path

here = Path(__file__).parent.absolute()


def test_all_modules_documented():
    docs_index = here / "../docs/index.rst"
    with open(docs_index) as fp:
        index_data = fp.read()

    ignored = {
        "__init__.py",
    }

    no_doc = []
    for mod in (here.parent / "autokitteh").glob("*.py"):
        if mod.name in ignored:
            continue

        ref = f".. automodule:: autokitteh.{mod.stem}"
        if ref not in index_data:
            no_doc.append(mod.name)

    if no_doc:
        no_doc = ", ".join(sorted(no_doc))
        assert False, f"missing from docs/index.rst: {no_doc}"



import sys
from pathlib import Path
from shutil import which
from subprocess import run
from os import environ

import pytest

uv_exe = which("uv")
here = Path(__file__).parent.absolute()
sdk_path = (here / "..").absolute()

py_code = """
from autokitteh import packages

{patch}

# click is small, pure Python and has no external depenecies
packages.install('click')
"""

install_cases = [
    # install command, patch
    pytest.param([uv_exe, "venv"], "", id="uv"),
    pytest.param(
        [sys.executable, "-m", "venv"],
        "packages.which = lambda v: None",
        id="pip",
    ),
]


@pytest.mark.parametrize("venv_cmd, patch", install_cases)
def test_install(venv_cmd, patch, tmp_path):
    if "uv" in venv_cmd[0] and not which("uv"):
        pytest.skip("uv not installed")

    venv_path = tmp_path / "venv"
    venv_cmd += [str(venv_path)]
    out = run(venv_cmd)
    assert out.returncode == 0

    py_script = tmp_path / "main.py"
    with open(py_script, "w") as fp:
        fp.write(py_code.format(patch=patch))

    # Make sure the script can import 'autokitteh'
    env = environ.copy()
    pypath = env.get("PYTHONPATH")
    env["PYTHONPATH"] = f"{sdk_path}:{pypath}"

    venv_py = venv_path / "bin/python"

    # If ak doesn't use uv, this will fail since there's no pip in venv created by uv
    out = run([str(venv_py), str(py_script)], env=env)
    assert out.returncode == 0



"""Initialize a Linear client, based on an AutoKitteh connection."""

import os
from requests import Session

from .connections import check_connection_name
from .oauth2_session import OAuth2Session


def linear_client(connection: str) -> Session:
    """Initialize a Linear client, based on an AutoKitteh connection.

    API reference:
    https://linear.app/developers/graphql

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Requests session object.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
    """
    check_connection_name(connection)
    api_key = os.getenv(f"{connection}__api_key")
    if api_key:
        session = Session()
        session.headers.update(
            {"Authorization": api_key, "Content-Type": "application/json"}
        )
        session.base_url = "https://api.linear.app/graphql/"
        return session

    return OAuth2Session(
        "linear", connection, base_url="https://api.linear.app/graphql/"
    )



"""AutoKitteh Python SDK."""

# Modules imported here should use only the standard library.
# We don't want people installing everything just to use autokitteh in testing

from . import errors
from .activities import activity, inhibit_activities, register_no_activity
from .attr_dict import AttrDict
from .event import Event
from .events import next_event, start, subscribe, unsubscribe
from .outcomes import outcome, http_outcome
from .signals import Signal, next_signal, signal
from .errors import AutoKittehError
from .store import (
    add_values,
    del_value,
    get_value,
    list_values_keys,
    mutate_value,
    set_value,
    store,
)

__all__ = [
    "AttrDict",
    "AutoKittehError",
    "errors",
    "http_outcome",
    "outcome",
    "start",
    # Activities
    "activity",
    "inhibit_activities",
    "register_no_activity",
    # Events
    "Event",
    "next_event",
    "subscribe",
    "unsubscribe",
    # Signals
    "next_signal",
    "signal",
    "Signal",
    # Values
    "add_values",
    "del_value",
    "get_value",
    "list_values_keys",
    "mutate_value",
    "set_value",
    "store",
]



import json
import os
from types import ModuleType

from kubernetes import client, config
from .connections import check_connection_name
from .errors import ConnectionInitError


def kubernetes_client(connection: str) -> ModuleType:
    """Initialize a Kubernetes client, based on an AutoKitteh connection.

    API reference:
    https://github.com/kubernetes-client/python
    https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.30/

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Kubernetes API client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: If the connection config is missing or invalid,
            or if an unexpected error occurs during client initialization.

    """
    check_connection_name(connection)

    config_file = os.getenv(connection + "__configFile")

    if not config_file:
        raise ConnectionInitError(connection)

    try:
        config_dict = json.loads(config_file)
        config.load_kube_config_from_dict(config_dict)
    except config.ConfigException as e:
        raise ConnectionInitError(connection) from e
    except Exception as e:
        raise RuntimeError(
            f"Internal error while initializing Kubernetes client for connection '{connection}''"
        ) from e

    return client



from collections.abc import MutableMapping
from typing import Any
from enum import StrEnum

# Dummy implementation for local development.
_local_dev_store = {}


class Store(MutableMapping):
    """Store it a dict like interface to ak store.

    Note that read-modify-write operations are not atomic.

    Values must be pickleable, see
    https://docs.python.org/3/library/pickle.html#what-can-be-pickled-and-unpickled

    Works both for durable and non-durable sessions.
    """

    def __getitem__(self, key):
        return get_value(key)

    def __setitem__(self, key, value):
        set_value(key, value)

    def __delitem__(self, key):
        del_value(key)

    def __iter__(self):
        return iter(list_values_keys())

    def __len__(self):
        return sum(1 for _ in self)


store = Store()


class Op(StrEnum):
    """Enum for operation types."""

    SET = "set"
    GET = "get"
    DEL = "del"
    ADD = "add"


def mutate_value(key: str, op: Op, *args: list[Any]) -> Any:
    """Mutate a stored value.

    Works both for durable and non-durable sessions.

    Args:
        key: Key of the value to mutate.
        op: Operation to perform on the value.
        args: Additional arguments for the operation.

    Returns:
        Any: Result of the operation, if applicable.

    Raises:
        AutoKittehError: Value is too large.
    """
    # Dummy implementation for local development.
    return {
        "set": set_value,
        "get": get_value,
        "del": del_value,
    }[op](key, *args)


def get_value(key: str) -> Any:
    """Get a stored value.

    Works both for durable and non-durable sessions.

    Args:
        key: Key of the value to retrieve.

    Returns:
        Any: The stored value, or None if not found.
    """

    # Dummy implementation for local development.
    return _local_dev_store.get(key)


def set_value(key: str, value: Any) -> None:
    """Set a stored value.

    Works both for durable and non-durable sessions.

    Args:
        key: Key of the value to set.
        value: Value to store. If Value is None, it will be deleted. Value must be serializable.

    Returns:
        None.

    Raises:
        AutoKittehError: Value is too large.
    """

    # Dummy implementation for local development.
    _local_dev_store[key] = value


def add_values(key: str, value: int | float) -> int | float:
    """Add to a stored value.

    This operation is atomic.

    If key is not found, its initial value is set to the provided value.

    Works both for durable and non-durable sessions.

    Args:
        key: Key of the value to set.
        value: Value to add. Value must be serializable.

    Returns:
        New result value. Always the same type as the value stored under the key.
    """

    # Dummy implementation for local development.
    _local_dev_store[key] = _local_dev_store.get(key, 0) + value
    return _local_dev_store[key]


def del_value(key: str) -> None:
    """Delete a stored value.

    Works both for durable and non-durable sessions.

    Args:
        key: Key of the value to set.

    Returns:
        None.
    """

    # Dummy implementation for local development.
    try:
        del _local_dev_store[key]
    except KeyError:
        pass


def list_values_keys() -> list[str]:
    """List all stored keys.

    Works both for durable and non-durable sessions.

    Returns:
        list[str]: Sorted list of all keys in the store.
    """

    # Dummy implementation for local development.
    return sorted(list(_local_dev_store.keys()))



"""Initializes a Reddit client, based on an AutoKitteh connection."""

import os

import praw

from .connections import check_connection_name
from .errors import ConnectionInitError


def reddit_client(connection: str) -> praw.Reddit:
    """Initialize a Reddit client, based on an AutoKitteh connection.

    API reference:
    https://praw.readthedocs.io/en/stable/

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Reddit API client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
    """
    check_connection_name(connection)

    client_id = os.getenv(connection + "__client_id")
    client_secret = os.getenv(connection + "__client_secret")
    user_agent = os.getenv(connection + "__user_agent")
    username = os.getenv(connection + "__username")
    password = os.getenv(connection + "__password")

    if not client_id or not client_secret or not user_agent:
        raise ConnectionInitError(connection)

    if username and password:
        return praw.Reddit(
            client_id=client_id,
            client_secret=client_secret,
            user_agent=user_agent,
            username=username,
            password=password,
        )

    # read-only client.
    return praw.Reddit(
        client_id=client_id,
        client_secret=client_secret,
        user_agent=user_agent,
    )



#!/usr/bin/env python

from pathlib import Path
from argparse import ArgumentParser, FileType

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "output",
    help="output requirements file",
    default="index.rst",
    type=FileType("w"),
    nargs="?",
)
args = parser.parse_args()

docs_dir = Path(__file__).parent.absolute()
ak_dir = docs_dir / "../autokitteh"

index_header = """\
.. autokitteh documentation master file, created by sphinx-quickstart & gen-index.py
   sphinx-quickstart on Wed Jul  3 11:27:22 2024.
   You can adapt this file completely to your liking, but it should at least
   contain the root `toctree` directive.

AutoKitteh Python SDK Documentation
===================================

.. toctree::
   :maxdepth: 2
   :caption: Contents:


Module contents
---------------

.. automodule:: autokitteh
   :members:
   :undoc-members:
   :show-inheritance:



Submodules
----------
"""

mod_header = "autokitteh.{mod} module"
mod_template = """\
{header}
{line}

.. automodule:: autokitteh.{mod}
   :members:
   :undoc-members:
   :show-inheritance:
"""

print(index_header, file=args.output)

ignored = {
    "__init__.py",
}
for mod in ak_dir.glob("*.py"):
    if mod.name in ignored:
        continue

    header = mod_header.format(mod=mod.stem)
    line = "-" * len(header)
    text = mod_template.format(mod=mod.stem, header=header, line=line)
    print(text, file=args.output)


footer = """\
Indices and tables
==================

* :ref:`genindex`
* :ref:`modindex`
* :ref:`search`
"""
print(footer, file=args.output)



from typing import Any


def outcome(v: Any) -> None:
    """Log an outcome for the current session.

    Works both in durable and nondurable sessions.

    Args:
        v: The outcome value. Can be any JSON-serializable value.

    """
    # Dummy implementation for local development.
    pass


def http_outcome(
    status_code: int = 200,
    body: Any = None,
    json: Any = None,
    headers: dict[str, str] = {},
    more: bool = False,
) -> None:
    """Respond to an HTTP request.

    Works both in durable and nondurable sessions.

    Args:
        status_code: HTTP status code to return. Ignored if not the first response to a request.
        body: body to return. If it is a dict or a list, it will be serialized as
            JSON. If it is a string or bytes, it will be returned as-is.
        json: JSON-serializable value to return as JSON. If specified, the Content-Type
            header will be set to application/json. Cannot be used together with body.
        headers: dict of headers to return.
        more: If True, indicates that more responses will follow for this request.
    """
    # Dummy implementation for local development.
    pass



from autokitteh import activities


def test_register_no_activity(monkeypatch):
    no_act = set()
    monkeypatch.setattr(activities, "_no_activity", no_act)

    fns = [dict.get, list.append]
    # Twice to make sure no duplicates
    activities.register_no_activity(fns)
    activities.register_no_activity(fns)

    assert no_act == set(fns)



"""Initialize an Airtable client, based on an AutoKitteh connection."""

import os

from pyairtable import Api

from .connections import check_connection_name, refresh_oauth
from .errors import ConnectionInitError, OAuthRefreshError


def airtable_client(connection: str) -> Api:
    """Initialize an Airtable client, based on an AutoKitteh connection.

    API reference:
    https://pyairtable.readthedocs.io/en/stable/getting-started.html
    https://github.com/gtalarico/pyairtable

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Requests session object.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        OAuthRefreshError: OAuth token refresh failed.
    """
    check_connection_name(connection)

    pat = os.getenv(f"{connection}__pat")
    if pat:
        return Api(pat)

    refresh_token = os.getenv(connection + "__oauth_refresh_token")
    if not refresh_token:
        raise ConnectionInitError(connection)

    try:
        access_token, _ = refresh_oauth("airtable", connection)
    except Exception as e:
        raise OAuthRefreshError(connection, str(e)) from e

    return Api(access_token)



"""Initialize an Anthropic client, based on an AutoKitteh connection."""

import os

from pipedrive.client import Client

from .connections import check_connection_name
from .errors import ConnectionInitError


def pipedrive_client(connection: str) -> Client:
    """Initialize a Pipedrive client, based on an AutoKitteh connection.

    API reference:
        https://pypi.org/project/pipedrive-python-lib/

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Pipedrive API client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
    """
    check_connection_name(connection)

    domain = os.getenv(connection + "__company_domain")
    api_key = os.getenv(connection + "__api_key")

    if not api_key or not domain:
        raise ConnectionInitError(connection)

    client = Client(domain=domain)
    client.set_api_token(api_key)
    return client



"""Initialize an Anthropic client, based on an AutoKitteh connection."""

import os

from anthropic import Anthropic

from .connections import check_connection_name
from .errors import ConnectionInitError


def anthropic_client(connection: str) -> Anthropic:
    """Initialize an Anthropic client, based on an AutoKitteh connection.

    API reference:
    https://docs.anthropic.com/claude/reference
    https://github.com/anthropics/anthropic-sdk-python

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Anthropic API client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        anthropic.APIError: Anthropic SDK initialization errors.
    """
    check_connection_name(connection)

    api_key = os.getenv(connection + "__api_key")

    if not api_key:
        raise ConnectionInitError(connection)

    return Anthropic(api_key=api_key)
